/* Copyright (C) 2003 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


#include <mysql.h>
#include <ndbapi/NdbApi.hpp>

// Used for cout
#include <stdio.h>
#include <iostream>
//#include <time.h>
#include <sys/time.h>

using namespace std; 

#define PRINT_ERROR(code,msg) \
  std::cout << "Error in " << __FILE__ << ", line: " << __LINE__ \
            << ", code: " << code \
            << ", msg: " << msg << "." << std::endl
#define MYSQLERROR(mysql) { \
  PRINT_ERROR(mysql_errno(&mysql),mysql_error(&mysql)); \
  exit(-1); }
#define APIERROR(error) { \
  PRINT_ERROR(error.code,error.message); \
  exit(-1); }

#define NUM_ITER 100000
//#define INSERT_NUM 100000000
#define INSERT_NUM 100000000
#define BATCH_SIZE 10000
#define BATCH_SLEEP true
#define BATCH_SLEEP_INTERVAL 10
#define BATCH_SLEEP_TIME 1
#define CREATE_TABLES true

int main()
{

  int default_int = 0; 
  Uint64 default_datetime = (Uint64)19000101*1000000+120000;
  char* default_varchar = "\020aaaaaaaaaaaaaaaaaaaa";


  ndb_init();


  /**************************************************************
   * Connect to mysql server and create table                   *
   **************************************************************/
  if (CREATE_TABLES) {
    MYSQL mysql;

    
    if ( !mysql_init(&mysql) ) {
      std::cout << "mysql_init failed\n";
      exit(-1);
    }
    mysql_options(&mysql,MYSQL_READ_DEFAULT_GROUP,"client");
    if ( !mysql_real_connect(&mysql, "localhost", "root", "", "",
			     3306, NULL, 0) )
      MYSQLERROR(mysql);

    std::cout << "purging db .. " << std::flush; 
    mysql_query(&mysql, "DROP DATABASE bulkload");
    mysql_query(&mysql, "CREATE DATABASE bulkload");
    std::cout << "done" << std::endl;

    if (mysql_query(&mysql, "USE bulkload") != 0) MYSQLERROR(mysql);

    std::cout << "creating schema ... " << std::flush; 
    if (mysql_query(&mysql, 
		    "create table kv_int ( "
		    "  id int not null auto_increment, "
		    "  val int not null, "
		    "  primary key (id) using hash"
		    ") engine=ndbcluster "
		    ))
      MYSQLERROR(mysql);

    if (mysql_query(&mysql, 
		    "create table kv_datetime ( "
		    "  id int not null auto_increment, "
		    "  val datetime not null, "
		    "  primary key (id) using hash"
		    ") engine=ndbcluster "
		    ))
      MYSQLERROR(mysql);

    if (mysql_query(&mysql, 
		    "create table kv_varchar ( "
		    "  id int not null auto_increment, "
		    "  val varchar(80) not null, "
		    "  primary key (id) using hash"
		    ") engine=ndbcluster "
		    ))
      MYSQLERROR(mysql);
    std::cout << "done" << std::endl;
  }


  /**************************************************************
   * Connect to ndb cluster                                     *
   **************************************************************/

  Ndb_cluster_connection *cluster_connection=
    new Ndb_cluster_connection("127.0.0.1"); // Object representing the cluster

  std::cout << "Connecting to cluster ... " << std::flush; 
  if (cluster_connection->connect(5,3,1))
  {
    std::cout << std::endl << "Connect to cluster management server failed.\n";
    exit(-1);
  }

  std::cout << "waiting ... " << std::flush; 
  if (cluster_connection->wait_until_ready(30,30))
  {
    std::cout << std::endl << "Cluster was not ready within 30 secs.\n";
    exit(-1);
  }
  std::cout << "ready" << std::endl; 

  Ndb* myNdb = new Ndb( cluster_connection,
			"bulkload" );  // Object representing the database
  if (myNdb->init() == -1) { 
    APIERROR(myNdb->getNdbError());
    exit(-1);
  }

  const NdbDictionary::Dictionary* myDict= myNdb->getDictionary();
  const NdbDictionary::Table *kv_int = myDict->getTable("kv_int");
  const NdbDictionary::Table *kv_datetime = myDict->getTable("kv_datetime");
  const NdbDictionary::Table *kv_varchar = myDict->getTable("kv_varchar");
  
  if (kv_int == NULL)
    APIERROR(myDict->getNdbError());
  if (kv_datetime == NULL)
    APIERROR(myDict->getNdbError());
  if (kv_varchar == NULL)
    APIERROR(myDict->getNdbError());


  int elapsed = 0;
  /**************************************************************************
   * Fill table with tuples, using auto_increment IDs                       *
   **************************************************************************/

  std::cout << "Filling tables" << std::endl << std::endl; 


  {
    timeval start_time; 
    timeval end_time; 
    gettimeofday(&start_time, NULL);
    double ustart = start_time.tv_sec+(start_time.tv_usec/1000000.0);

//    std::time_t before_t = std::time(0);
    for (int f = 0; f < INSERT_NUM/BATCH_SIZE; f++) {
      NdbTransaction *myTransaction= myNdb->startTransaction();
      if (myTransaction == NULL) APIERROR(myNdb->getNdbError());
      
      for (int i = 1; i<= BATCH_SIZE; i ++ ) { 


	NdbOperation *intOperation = myTransaction->getNdbOperation(kv_int);
	if (intOperation == NULL) 
	  APIERROR(myTransaction->getNdbError());
	
	intOperation->insertTuple();
	
	if (intOperation->equal("id",(int)i+(f*BATCH_SIZE))==-1)
	  APIERROR(myTransaction->getNdbError());
	
	if (intOperation->setValue("val",default_int)==-1)
	  APIERROR(myTransaction->getNdbError());

	NdbOperation *varcharOperation = myTransaction->getNdbOperation(kv_varchar);
	if (varcharOperation == NULL) 
	  APIERROR(myTransaction->getNdbError());
	
	varcharOperation->insertTuple();
	
	if (varcharOperation->equal("id",(int)i+(f*BATCH_SIZE))==-1)
	  APIERROR(myTransaction->getNdbError());
	
	if (varcharOperation->setValue("val",default_varchar)==-1)
	  APIERROR(myTransaction->getNdbError());

	NdbOperation *datetimeOperation = myTransaction->getNdbOperation(kv_datetime);
	if (datetimeOperation == NULL) 
	  APIERROR(myTransaction->getNdbError());
	
	datetimeOperation->insertTuple();
	
	if (datetimeOperation->equal("id",(int)i+(f*BATCH_SIZE))==-1)
	  APIERROR(myTransaction->getNdbError());
	
	if (datetimeOperation->setValue("val",default_datetime)==-1)
	  APIERROR(myTransaction->getNdbError());

      }

      if (myTransaction->execute( NdbTransaction::Commit ) == -1)
	APIERROR(myTransaction->getNdbError());
      
      myNdb->closeTransaction(myTransaction);

gettimeofday(&end_time, NULL);
	int this_elapsed = end_time.tv_sec - start_time.tv_sec; 
	if (this_elapsed > elapsed ) {
	  elapsed = this_elapsed; 
	  std::cout << "Inserted " << (f+1)*BATCH_SIZE << " ";
	  std::cout << "(" << (f+1)*BATCH_SIZE*3 << " records) ";
	  std::cout << elapsed << std::endl;
	}
     // }
     if (BATCH_SLEEP)
     	if (f%BATCH_SLEEP_INTERVAL==0)
	  sleep(BATCH_SLEEP_TIME);

    }
    
    gettimeofday(&end_time, NULL);
    double uend = end_time.tv_sec+(end_time.tv_usec/1000000.0);
    
    std::cout << "Insert time for " << INSERT_NUM << " transactions ";
    std::cout << "(" << INSERT_NUM*3 << " records) ";
    std::cout << " = " << uend-ustart << std::endl;
    
   
  }


  delete myNdb;
  delete cluster_connection;

  ndb_end(0);
  return 0;
}
